@@ -8,7 +8,8 @@ import typing
 import requests
 import pytz
 
-from . import api, fields, records, utils
+from . import api, fields, utils
+from . import record as knackpy_record
 from .models import TIMEZONES
 
 
@@ -50,8 +51,8 @@ class App:
         slug: str = None,
         metadata: str = None,
         tzinfo: datetime.tzinfo = None,
-        max_attempts: int = None,
-        timeout: int = None,
+        max_attempts: int = 5,
+        timeout: int = 30,
     ):
 
         if not api_key:
@@ -76,6 +77,7 @@ class App:
         self.field_defs = fields.field_defs_from_metadata(self.metadata)
         self.containers = utils.generate_containers(self.metadata)
         self.data = {}
+        self.records = {}
         logging.debug(self)
 
     def _get_metadata(self):
@@ -179,29 +181,13 @@ class App:
                 f"Unknown container specified: {identifier}. Inspect App.containers for available containers."  # noqa
             )
 
-    def _build_request_kwargs(self, **kwargs) -> dict:
-        """Compile the keyword arguments to be passed to `knackpy.api`. We drop params
-        that are NoneType because we don't want to override the default values for
-        these params that are define in the api methods.
-
-        Args:
-            record_limit (int): the maximum number of records to retrieve.
-            max_attempts (int): The maximum number of attempts to make if a request
-                times out.
-            timeout (int, optional): Number of seconds to wait before a Knack API
-                request times out. Further reading:
-                [Requests docs](https://requests.readthedocs.io/en/master/user/quickstart/).  # noqa:E501
-        """
-        supported_kwargs = ["record_limit", "max_attempts", "timeout"]
-
-        return {key: kwargs[key] for key in supported_kwargs if kwargs.get(key)}
-
-    def records(
+    def get(
         self,
         identifier: str = None,
         refresh: bool = False,
         record_limit: int = None,
         filters: typing.Union[dict, list] = None,
+        generate=False,
     ):
         """Get records from a knack object or view.
 
@@ -220,6 +206,8 @@ class App:
                     `None`, will return all records.
                 filters (dict or list, optional): A dict or of Knack API filiters.
                     See: https://www.knack.com/developer-documentation/#filters.
+                generate (bool, optional): If True, will return a generator which
+                    yields knacky.Record objects instead of return a list of of them.
 
             Returns:
                 A `generator` which yields knackpy Record objects.
@@ -231,15 +219,18 @@ class App:
 
         container = self._find_container(identifier)
 
+        # note that data is always assigned to an object or view key, regardless of
+        # whether or not the client provides an object or view *name*
         container_key = container.obj or container.view
 
-        if not self.data.get(container_key) or refresh:
-            request_kwargs = self._build_request_kwargs(
-                max_attempts=self.max_attempts,
-                timeout=self.timeout,
-                record_limit=record_limit,
-            )
+        if self.records.get(container_key) and not refresh:
+            # if the data has already been retrieved we do not fetch it again or convert
+            # the data into knackpy.record.Record's again, unless refresh.
+            return self.records[container_key]
 
+        if not self.data.get(container_key) or refresh:
+            # to support side-loading of data, we may have a situation where data is
+            # present, but records have not never been generated
             self.data[container_key] = api.get(
                 app_id=self.app_id,
                 api_key=self.api_key,
@@ -248,15 +239,52 @@ class App:
                 view=container.view,
                 filters=filters,
                 slug=self.slug,
-                **request_kwargs,
+                max_attempts=self.max_attempts,
+                timeout=self.timeout,
+                record_limit=record_limit,
             )
 
-        return self._generate_records(container_key, self.data[container_key])
+        self.records[container_key] = self._records(container_key, generate)
+        return self.records[container_key]
 
-    def _generate_records(self, container_key, data):
-        return records.Records(
-            container_key, data, self.field_defs, self.timezone
-        ).records()
+    def _records(self, container_key, generate=False):
+        """Return a list or generator of knackpy.record.Record objects.
+
+        Args:
+            container_key (str): An Knack object or view key.
+            generate (bool, optional): If true, will return a Record generator function
+                instead of a list of Record's.
+
+        Returns:
+            list or generator: A list or generator of knackpy.record.Record's.
+        """
+        data = self.data[container_key]
+
+        # filter field defs by requested container
+        field_defs = [
+            field_def
+            for field_def in self.field_defs
+            if container_key == field_def.obj or container_key in field_def.views
+        ]
+
+        try:
+            identifier = [
+                field_def.key for field_def in field_defs if field_def.identifier
+            ][0]
+        except IndexError:
+            identifier = None
+
+        if generate:
+            return self._generate_records(data, field_defs, identifier)
+
+        return [
+            knackpy_record.Record(record, field_defs, identifier, self.timezone)
+            for record in data
+        ]
+
+    def _generate_records(self, data, field_defs, identifier):
+        for record in data:
+            yield knackpy_record.Record(record, field_defs, identifier, self.timezone)
 
     def _find_field_def(self, identifier, obj):
         return [
@@ -279,7 +307,7 @@ class App:
         if not os.path.exists(out_dir):
             os.makedirs(out_dir)
 
-        records = self.records(identifier)
+        records = self.get(identifier)
 
         csv_data = [record.format() for record in records]
 
@@ -328,7 +356,7 @@ class App:
 
         downloads = []
 
-        for record in self.records(identifier):
+        for record in self.get(identifier):
             file_dict = record.raw.get(field_key_raw)
 
             if not file_dict:
@@ -390,8 +418,8 @@ class App:
 
     def download(
         self,
-        identifier: str,
         *,
+        container: str,
         field: str,
         out_dir: str = "_downloads",
         label_keys: list = None,
@@ -399,7 +427,7 @@ class App:
         """Download files and images from Knack records.
 
         Args:
-            identifier (str): The name or key of the object from which files will be
+            container (str): The name or key of the object from which files will be
                 downloaded.
             out_dir (str, optional): Relative path to the directory to which files
                 will be written. Defaults to "_downloads".
@@ -414,15 +442,15 @@ class App:
         if not os.path.exists(out_dir):
             os.makedirs(out_dir)
 
-        container = self._find_container(identifier)
+        download_container = self._find_container(container)
 
-        field_defs = self._find_field_def(field, identifier)
+        field_defs = self._find_field_def(field, container)
 
         if not field_defs:
             raise ValueError(f"Field not found: '{field}'")
 
         downloads = self._assemble_downloads(
-            container.obj, field_defs[0].key, label_keys, out_dir
+            download_container.obj, field_defs[0].key, label_keys, out_dir
         )
 
         download_count = self._download_files(downloads)
@@ -430,3 +458,119 @@ class App:
         logging.debug(f"{download_count} files downloaded.")
 
         return download_count
+
+    def _replace_record(self, record, obj):
+        for rec in self.data[obj]:
+            if rec["id"] == record["id"]:
+                rec = record
+                return self.data[obj]
+
+    def _update_record_state(self, res, obj, method, record_id=None):
+        """ Keep local data and records in sync with CRUD operations.
+
+        Args:
+            res (dict): Knack API response. Either a record `dict` or `{"delete": True}`
+            obj (str): The Knack object key that was updated.
+            method (str): `create`, `update`, or `delete`.
+            record_id (str): The Knack record ID of the affected record, if applicable.
+
+        Side-Effects:
+            Update `self.data[key]` accordingly; regenerate `self.records[key]`.
+
+        Returns:
+            None
+        """
+        if method == "add":
+            self.data[obj].append(res)
+
+        elif method == "update":
+            self.data[obj] = self._replace_record(res, obj)
+
+        elif method == "delete":
+            # the Knack API responds with {"delete": True} when deleting records so we
+            # need to have the record_id of the deleted record explicitly here (ie
+            # there is no `record` response to work with)
+            self.data[obj] = [
+                record for record in self.data[obj] if record["id"] != record_id
+            ]
+
+        self.records[obj] = self._records(obj)
+        return None
+
+    def record(
+        self, *, data: dict, method: str, obj: str,
+    ):
+        """Create, update, or delete a Knack record.
+
+        Args:
+            data (dict): The Knack record data payload.
+            method (str): Choose from `create`, `update`, or `delete`.
+            obj (str, optional): The Knack object key or name which holds the record
+                data.
+
+        Returns:
+            dict: The updated or newly created Knack record data, or, if deleting a
+                record: `{"delete": true}`
+        """
+
+        # if the client provides a view identifier, it will raise no error until the
+        # Knack API raises an HTTPError
+        container = self._find_container(obj)
+
+        res = api.record(
+            app_id=self.app_id,
+            api_key=self.api_key,
+            data=data,
+            method=method,
+            obj=container.obj,
+            slug=self.slug,
+            max_attempts=self.max_attempts,
+            timeout=self.timeout,
+        )
+
+        if self.data.get(obj):
+            # if data for the affected obj is stored locally, update it accordingly.
+            self._update_record_state(res, obj, method, record_id=data.get("id"))
+        return res
+
+    def upload(
+        self,
+        *,
+        container: str,
+        field: str,
+        path: str,
+        asset_type: str,
+        record_id: str = None,
+    ):
+        """Upload a file or image to Knack. This is a two-step process:
+
+        1) Upload file asset to Knack storage
+        2) Create/update a record that links to the file in storage
+
+        Knack docs: https://www.knack.com/developer-documentation/#file-image-uploads
+
+        Args:
+            container (str): The name or key of the object from which files will be
+                downloaded.
+            field (str): The knack field key of the field you're uploading into.
+            path (str): The path to the file to be uploaded.
+            asset_type (str): The type of Knack field you're uploading to. Must be
+                `file` or `image`.
+            record_id (str, optional): The knack record ID to which the upload will be
+                attached. If `None`, will create a new record. Otherwise will update an
+                existing record.
+        """
+        download_container = self._find_container(container)
+
+        return api.upload(
+            app_id=self.app_id,
+            api_key=self.api_key,
+            obj=download_container.obj,
+            field=field,
+            path=path,
+            asset_type=asset_type,
+            record_id=record_id,
+            slug=self.slug,
+            max_attempts=self.max_attempts,
+            timeout=self.timeout,
+        )
